
/* Language C Like String Formatter
 * 
 * Copyright (C) 2008 S.Fuchita.
 * http://www.soramimi.jp/
 */

package jp.soramimi.text;

import java.io.PrintStream;
import java.lang.reflect.Field;
import java.util.Vector;


public final class Format {

	private static String _digit_lower = "0123456789abcdef";
	private static String _digit_upper = "0123456789ABCDEF";

	private String _format;
	private StringBuffer _output;

	private Vector _vector = new Vector();
	private int _argpos;

	private int _width;
	private int _precision;
	private boolean _leadzero;
	private boolean _leftalign;
	private int _count;
	private char _plus;
	private char _minus;

	private void PUTC(char c)
	{
		_output.append(c);
	}
	private void PUTN(char c, int n)
	{
		while (n > 0) {
			_output.append(c);
			n--;
		}
	}
	private void PUTS(String s)
	{
		_output.append(s);
	}


	private static double log10(double v)
	{
		return Math.log(v) / Math.log(10);
	}


	private void _format_du(boolean sign, long v)
	{
		int i, len, size, width, precision;
		byte[] tmp = new byte [20];

		char signchar = sign ? _minus : _plus;

		width = _width;
		precision = _precision;
		if (_leadzero && precision < 0 && !_leftalign) {
			precision = width;
			if (signchar != 0) precision--;
		}
		if (precision < 1) precision = 1;


		tmp[0] = (byte)(v & 1);
		v = (v >> 1) & 0x7fffffffffffffffL;
		tmp[0] += (byte)(v % 5) * 2;
		v /= 5;
		len = 1;
		while (v != 0) {
			tmp[len] = (byte)(v % 10);
			v /= 10;
			len++;
		}

		size = len < precision ? precision : len;
		if (signchar != 0) size++;

		if (!_leftalign && width > size) {
			PUTN(' ', width - size);
		}
		if (signchar != 0) {
			PUTC(signchar);
		}
		for (i = len; i < precision; i++) {
			PUTC('0');
		}
		while (len > 0) {
			PUTC(_digit_lower.charAt(tmp[--len]));
		}
		if (_leftalign && width > size) {
			PUTN(' ', width - size);
		}
	}
	private void _format_x(long value, boolean upper)
	{
		int i, len, size, width, precision;
		byte[] tmp = new byte [100];

		String hex = upper ? _digit_upper : _digit_lower;


		width = _width;
		precision = _precision;
		if (_leadzero && precision < 0 && !_leftalign) {
			precision = width;
		}
		if (precision < 1) precision = 1;

		tmp[0] = (byte)(value & 15);
		value = (value >> 4) & 0x0fffffffffffffffL;
		len = 1;
		while (value != 0) {
			tmp[len] = (byte)(value & 15);
			value >>= 4;
			len++;
		}
		size = len < precision ? precision : len;

		if (!_leftalign && width > size) {
			PUTN(' ', width - size);
		}
		for (i = len; i < precision; i++) {
			PUTC('0');
		}
		while (len > 0) {
			PUTC(hex.charAt(tmp[--len]));
		}
		if (_leftalign && width > size) {
			PUTN(' ', width - size);
		}
	}
	private void _format_o(long value)
	{
		int i, len, size, width, precision;
		byte[] tmp = new byte [100];

		width = _width;
		precision = _precision;
		if (_leadzero && precision < 0 && !_leftalign) {
			precision = width;
		}
		if (precision < 1) precision = 1;

		tmp[0] = (byte)(value & 7);
		value = (value >> 3) & 0x1fffffffffffffffL;
		len = 1;
		while (value != 0) {
			tmp[len] = (byte)(value & 7);
			value >>= 3;
			len++;
		}
		size = len < precision ? precision : len;

		if (!_leftalign && width > size) {
			PUTN(' ', width - size);
		}
		for (i = len; i < precision; i++) {
			PUTC('0');
		}
		while (len > 0) {
			PUTC(_digit_lower.charAt(tmp[--len]));
		}
		if (_leftalign && width > size) {
			PUTN(' ', width - size);
		}
	}
	private void _format_b(long value)
	{
		int i, len, size, width, precision;
		long bits;

		width = _width;
		precision = _precision;
		if (_leadzero && precision < 0 && !_leftalign) {
			precision = width;
		}
		if (precision < 1) precision = 1;

		bits = value & 1;
		value = (value >> 1) & 0x7fffffffffffffffL;
		len = 1;
		while (value != 0) {
			bits = (bits << 1) | (value & 1);
			value >>= 1;
			len++;
		}
		size = len < precision ? precision : len;

		if (!_leftalign && width > size) {
			PUTN(' ', width - size);
		}
		for (i = len; i < precision; i++) {
			PUTC('0');
		}
		while (len > 0) {
			PUTC((char)('0' + (int)(bits & 1)));
			bits >>= 1;
			len--;
		}
		if (_leftalign && width > size) {
			PUTN(' ', width - size);
		}
	}
	private void _format_d(long value)
	{
		if (value < 0) {
			_format_du(true, -value);
		} else {
			_format_du(false, value);
		}
	}
	private void _format_u(long value)
	{
		_format_du(false, value);
	}
	private void _format_g(double value, int type)
	{
		int e, i, n, w, len, precision;
		char signchar;
		byte etype;
		byte[] tmp = new byte [100];

		signchar = _plus;
		if (value < 0) {
			signchar = _minus;
			value = -value;
		}

		precision = _precision;
		if (precision < 0) precision = 6;

		if (type == 'g' || type == 'G') {
			precision--;
			e = (int)Math.floor(log10(value));
			value = Math.floor(value * Math.pow(10, precision - e) + 0.5);

			n = 0;
			if (e >= 0 && e < precision) {
				n = e;
			}
			while (precision > n && (value % 10 == 0)) {
				value /= 10;
				precision--;
			}
			if (e >= -4 && e <= precision) {
				precision -= e;
				e = 0;
			}
		} else {
			e = 0;
			if (type == 'e' || type == 'E') {
				e = (int)Math.floor(log10(value));
			}
			value = Math.floor(value * Math.pow(10, precision - e) + 0.5);
		}

		len = 0;
		while (len < tmp.length) {
			tmp[len++] = (byte)(value % 10);
			value = Math.floor(value / 10);
			if (value == 0) break;
		}
		n = precision + 1;
		if (n < len) n = len;

		w = n;
		if (signchar != 0) w++;
		if (precision != 0) w++;

		i = 0;
		etype = 0;
		switch (type) {
		case 'G':	i++;
		case 'g':	i++;
		case 'E':	i++;
		case 'e':
			if (e == 0 && ((i & 2) != 0)) break;		// g, G
			etype = (byte)(((i & 1) != 0) ? 'E' : 'e');
			w += 5;
			break;
		}

		if (!_leftalign && w < _width) {
			if (_leadzero) {
				n += _width - w;
			} else {
				PUTN(' ', _width - w);
			}
		}

		if (signchar != 0) {
			PUTC(signchar);
		}

		while (n > 0) {
			if (n == precision) {
				PUTC('.');
			}
			if (--n < len) {
				PUTC(_digit_lower.charAt(tmp[n]));
			} else {
				PUTC('0');
			}
		}

		if (etype != 0) {
			PUTC((char)etype);
			if (e < 0) {
				PUTC('-');
				e = -e;
			} else {
				PUTC('+');
			}
			PUTC(_digit_lower.charAt(e / 100 % 10));
			PUTC(_digit_lower.charAt(e / 10 % 10));
			PUTC(_digit_lower.charAt(e % 10));
		}

		if (_leftalign && w < _width) {
			PUTN(' ', _width - w);
		}
	}
	private void pad(int n, boolean leadzero)
	{
		if (n > 0) PUTN(leadzero ? '0' : ' ', n);
	}
	private void _format_s(String value)
	{
		int n = value.length();
		n = (_width > n) ? (_width - n) : 0;

		if (!_leftalign) pad(n, _leadzero);
		PUTS(value);
		if (_leftalign) pad(n, false);
	}
	private void _format_c(char value)
	{
		int n = 1;
		n =(_width > n) ? (_width - n) : 0;

		if (!_leftalign) pad(n, _leadzero);
		if (value != 0) PUTC(value);
		if (_leftalign) pad(n, false);
	}

	private static String escapeString(String str, boolean singlequote, boolean doublequote)
	{
		int i, n;
		char c;
		StringBuffer sb = new StringBuffer();
		for (i = 0; i < str.length(); i++) {
			c = str.charAt(i);
			switch (c) {
			case '\b':	sb.append("\\b");	break;
			case '\t':	sb.append("\\t");	break;
			case '\n':	sb.append("\\n");	break;
			case '\f':	sb.append("\\f");	break;
			case '\r':	sb.append("\\r");	break;
			case '\'':	sb.append(singlequote ? "\\\'" : "'");	break;
			case '\"':	sb.append(doublequote ? "\\\"" : "\"");	break;
			case '\\':	sb.append("\\\\");	break;
			default:
				if (c < 0x20) {
					sb.append(new Format("\\x%02x")._(c).sprintf());
				} else {
					sb.append(c);
				}
				break;
			}
		}
		return new String(sb);
	}
	private void _format_p(Value value)
	{
		if (value != null) {
			_output.append(value.getDescription() + "\n");
		}
	}


	abstract private static class Value extends Number {
		public int intValue() { return 0; }
		public long longValue() { return 0; }
		public float floatValue() { return (float)doubleValue(); }
		public double doubleValue() { return 0; }
		public String toString() { return ""; }
		public char charValue() { return 0; }
		public String getDescription() { return "{" + toString() + "}"; }
	}
	private static final class VInteger extends Value {
		private int _value;
		VInteger(int v) { _value = v; }
		public int intValue() { return _value; }
		public long longValue() { return _value; }
		public double doubleValue() { return _value; }
		public String toString() { return Integer.toString(_value); }
		public char charValue() { return (char)_value; }
		public String getDescription() { return "{int " + toString() + "}"; }
	}
	private static final class VLong extends Value {
		private long _value;
		VLong(long v) { _value = v; }
		public int intValue() { return (int)_value; }
		public long longValue() { return _value; }
		public double doubleValue() { return _value; }
		public String toString() { return Long.toString(_value); }
		public char charValue() { return (char)_value; }
		public String getDescription() { return "{long " + toString() + "}"; }
	}
	private static final class VDouble extends Value {
		private double _value;
		VDouble(double v) { _value = v; }
		public int intValue() { return (int)_value; }
		public long longValue() { return (long)_value; }
		public double doubleValue() { return _value; }
		public String toString() { return Double.toString(_value); }
		public char charValue() { return (char)_value; }
		public String getDescription() { return "{double " + toString() + "}"; }
	}
	private static final class VCharacter extends Value {
		private char _value;
		VCharacter(char v) { _value = v; }
		public int intValue() { return _value; }
		public long longValue() { return _value; }
		public double doubleValue() { return _value; }
		public String toString() { return Character.toString(_value); }
		public char charValue() { return _value; }
		public String getDescription() { return "{char " + toString() + "}"; }
	}
	private static class VObject extends Value {
		protected Object _object;
		VObject(Object o) { _object = o; }
		public String getDescription()
		{
			if (_object == null) {
				return "{null}";
			}

			Class classString = null;
			Class classCharacter = null;
			try {
				classString = "".getClass();
				classCharacter = Class.forName("java.lang.Character");
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			}

			if (classString.isInstance(_object)) {
				return new Format("{%s \"%s\"}")
					._(_object.getClass().getName())
					._(escapeString(toString(), false, true))
					.sprintf();
			}

			Class c = _object.getClass();
			StringBuffer sb = new StringBuffer();

			sb.append(new Format("{%s \n")
				._(c.getName())
				.sprintf());

			Field[] f = c.getDeclaredFields();

			int i;
			for (i = 0; i < f.length; i++) {
				Class t = f[i].getType();
				Object v = null;
				try {
					v = f[i].get(_object);
					sb.append(new Format("\t%s %s = ")
						._(t.getName())
						._(f[i].getName())
						.sprintf());
					if (classCharacter.isInstance(v)) {
						String s = escapeString(v.toString(), true, false);
						sb.append(new Format("0x%04x;\t// \"%s\"")
							._(((Character)v)
							.charValue())
							._(s)
							.sprintf());
					} else if (t.isPrimitive()) {
						sb.append(v.toString() + ";");
					} else if (classString.isInstance(v)) {
							sb.append("\"" + escapeString(v.toString(), false, true) + "\";");
					} else {
						sb.append("{" + v.toString() + "};");
					}
					sb.append("\n");
				} catch (IllegalAccessException e) {
				}
			}
			sb.append("}");
			return new String(sb);
		}
	}
	private static final class VString extends VObject {
		VString(String s)
		{
			super(s);
		}
		private static double parseDouble(String str)
		{
			double v;
			char c;
			int i, n, e;

			v = 0;
			e = 0;
			c = 0;
			i = 0;
			while (i < str.length()) {
				c = str.charAt(i);
				if (e == 0) {
					if (c == '.') {
						e = 1;
						i++;
						continue;
					}
				}
				if (!Character.isDigit(c)) break;
				v = v * 10 + (c - '0');
				if (e > 0) e++;
				i++;
			}
			e = (e > 0) ? (1 - e) : 0;
			if (c == 'e' || c == 'E') {
				boolean sign = false;
				if (++i < str.length()) {
					c = str.charAt(i);
					switch (c) {
					case '+':
						i++;
						break;
					case '-':
						sign = true;
						i++;
						break;
					}
				}
				n = 0;
				while (i < str.length()) {
					c = str.charAt(i);
					if (!Character.isDigit(c)) break;
					n = n * 10 + (c - '0');
					i++;
				}
				e += sign ? -n : n;
			}
			if (e != 0) v *= Math.pow(10, e);
			return v;
		}
		public char charValue()
		{
			if (_object != null) {
				String s = _object.toString();
				if (s != null && s.length() > 0) {
					return s.charAt(0);
				}
			}
			return 0;
		}
		public int intValue() { return (int)doubleValue(); }
		public long longValue() { return (long)doubleValue(); }
		public float floatValue() { return (float)doubleValue(); }
		public double doubleValue() { return parseDouble(toString()); }
		public String toString() { return _object.toString(); }
//		public String getDescription() { return "{java.lang.String \"" + escapeString(toString(), false, true) + "\"}"; }
	}

	private Value getNextObject()
	{
		if (_argpos < _vector.size()) {
			return (Value)_vector.elementAt(_argpos++);
		}
		return null;
	}
	private Value getNextNumber()
	{
		return (Value)getNextObject();
	}
	private int getNextInt()
	{
		Value v = getNextNumber();
		if (v == null) return 0;
		return v.intValue();
	}
	private long getNextLong()
	{
		Value v = getNextNumber();
		if (v == null) return 0;
		return v.longValue();
	}
	private double getNextDouble()
	{
		Value v = getNextNumber();
		if (v == null) return 0;
		return v.doubleValue();
	}
	private char getNextChar()
	{
		Value v = getNextNumber();
		if (v == null) return 0;
		return (char)v.intValue();
	}
	private String getNextString()
	{
		Value v = getNextNumber();
		if (v == null) return "";
		return v.toString();
	}


	public void format(StringBuffer output)
	{
		_output = output;

		int pos;
		byte state;
		char c;
		_count = 0;
		state = 0;
		pos = 0;
		while (pos < _format.length()) {
			if (state == (byte)'%') {
				c = _format.charAt(pos++);
				if (c == '0' && _width < 0) {
					_leadzero = true;
				} else if (Character.isDigit(c)) {
					int n = c - '0';
					if (_width < 0) {
						_width = n;
					} else if (_precision < 0) {
						_width = _width * 10 + n;
					} else {
						_precision = _precision * 10 + n;
					}
				} else if (c == '-' && _width < 0) {
					_leftalign = true;
				} else if (c == '+' && _width < 0) {
					_plus = '+';
				} else if (c == ' ' && _width < 0) {
					_plus = ' ';
				} else if (c == '.') {
					if (_width < 0) {
						_width = 0;
					}
					if (_precision < 0) {
						_precision = 0;
					}
				} else if (c == '*') {
					if (_width <= 0) {
						_width = getNextInt();
					} else if (_precision == 0) {
						_precision = getNextInt();
					} else {
						PUTC(c);
						state = 0;
					}
				} else {
					switch (c) {
					case 's':	_format_s(getNextString());				break;
					case 'd':	_format_d(getNextLong());				break;
					case 'i':	_format_d(getNextLong());				break;
					case 'u':	_format_u(getNextLong());				break;
					case 'c':	_format_c(getNextChar());				break;
					case 'x':	_format_x(getNextLong(), false);		break;
					case 'X':	_format_x(getNextLong(), true);			break;
					case 'o':	_format_o(getNextLong());				break;
					case 'b':	_format_b(getNextLong());				break;
					case 'f':	_format_g(getNextDouble(), 'f');		break;
					case 'e':	_format_g(getNextDouble(), 'e');		break;
					case 'E':	_format_g(getNextDouble(), 'E');		break;
					case 'g':	_format_g(getNextDouble(), 'g');		break;
					case 'G':	_format_g(getNextDouble(), 'G');		break;
					case 'p':	_format_p(getNextObject());				break;
					default:
						PUTC(c);
						break;
					}
					state = 0;
				}
			} else {
				while (pos < _format.length()) {
					c = _format.charAt(pos++);
					if (c == '%') {
						state = (byte)c;
						_width = -1;
						_precision = -1;
						_leftalign = false;
						_leadzero = false;
						_plus = 0;
						_minus = '-';
						break;
					}
					PUTC(c);
				}
			}
		}
	}
	public String sprintf()
	{
		if (_output == null) {
			format(new StringBuffer());
		}
		return new String(_output);
	}
	public Format fprintf(PrintStream out)
	{
		out.print(sprintf());
		return this;
	}
	public Format printf()
	{
		return fprintf(System.out);
	}


	public Format(String str)
	{
		_format = str;
	}

	private void add(Object v)
	{
		_vector.add(v);
		_output = null;
	}

	public Format _(int value)
	{
		add(new VInteger(value));
		return this;
	}
	public Format _(long value)
	{
		add(new VLong(value));
		return this;
	}
	public Format _(double value)
	{
		add(new VDouble(value));
		return this;
	}
	public Format _(char value)
	{
		add(new VCharacter(value));
		return this;
	}
	public Format _(String value)
	{
		add(new VString(value));
		return this;
	}
	public Format _(Object value)
	{
		add(new VObject(value));
		return this;
	}

	public Format clear()
	{
		_vector.clear();
		_output = null;
		return this;
	}

	public static void dump(byte[] data)
	{
		int i, n, addr, max;
		max = 16;
		addr = 0;
		while (addr < data.length) {
			n = data.length - addr;
			if (n > max) n = max;
			new Format("%08X ")._(addr).printf();
			for (i = 0; i < max; i++) {
				if (i < n) {
					int c = data[addr + i] & 0xff;
					new Format("%02X ")._(c).printf();
				} else {
					System.out.print("   ");
				}
			}
			System.out.print("|");
			for (i = 0; i < n; i++) {
				int c = data[addr + i] & 0xff;
				if (c < 0x20 || c > 0x7f) c = ' ';
				new Format("%c")._(c).printf();
			}
			System.out.println("|");
			addr += n;
		}
	}
}



